from dataclasses import dataclass
import math

from .primitives import Angle, AngularSpeed, Au, AuSpeed
from .zodiac import Zodiac, ZodiacConstell


@dataclass
class EclCoord:
    """Ecliptic coordinate"""

    latitude: Angle
    longitude: Angle
    distance: Au

    def __init__(self, longitude: float, latitude: float, distance: float):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)
        self.distance = Au(distance)

    def __eq__(self, other) -> bool:
        return self.latitude == other.latitude and self.longitude == other.longitude

    def __xor__(self, other) -> Angle:
        if self == other:
            return Angle(0)
        long1 = self.longitude.radians()
        long2 = other.longitude.radians()
        lat1 = self.latitude.radians()
        lat2 = other.latitude.radians()
        dalpha = long1 - long2
        cosd = math.sin(lat1) * math.sin(lat2) + math.cos(lat1) * math.cos(lat2) * math.cos(dalpha)
        ang = math.acos(cosd)
        if ang > math.pi:
            ang = 2.0 * math.pi - ang
        return Angle(radians=ang)

    def sign_pos(self) -> (Zodiac, float):
        sign = Zodiac.from_longitude(self.longitude)
        return sign, self.longitude.degrees - sign.offset

    def constell_pos(self) -> (ZodiacConstell, float):
        constell = ZodiacConstell.from_longitude(self.longitude)
        pos = self.longitude.degrees
        if constell.upto_lng < constell.from_lng < pos:
            offset = pos - constell.from_lng
        elif pos < constell.upto_lng < constell.from_lng:
            offset = pos
        else:
            offset = pos - constell.from_lng
        return constell, offset

    def json(self) -> dict:
        return {'lat': self.latitude.degrees, 'long': self.longitude.degrees, 'dist': self.distance.au}


@dataclass
class EquatorCoord:
    """Equatorial coordinate"""

    ra: Angle
    decl: Angle
    distance: Au

    def __init__(self, ra: float, decl: float, distance: float):
        self.ra = Angle(ra)
        self.decl = Angle(decl)
        self.distance = Au(distance)

    def json(self) -> dict:
        return {'ra': self.ra.degrees, 'decl': self.decl.degrees, 'dist': self.distance.au}

@dataclass
class BaryCoord:
    """Barycentric coordinate"""

    latitude: Angle
    longitude: Angle
    distance: Au

    def __init__(self, longitude: float, latitude: float, distance: float):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)
        self.distance = Au(distance)

    def json(self) -> dict:
        return {'lat': self.latitude.degrees, 'long': self.longitude.degrees, 'dist': self.distance.au}

@dataclass
class HelioCoord:
    """Heliocentric coordinate"""

    latitude: Angle
    longitude: Angle
    distance: Au

    def __init__(self, longitude: float, latitude: float, distance: float):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)
        self.distance = Au(distance)

    def json(self) -> dict:
        return {'lat': self.latitude.degrees, 'long': self.longitude.degrees, 'dist': self.distance.au}


@dataclass
class HorCoord:
    """Horizontal coordinate"""

    azimuth: Angle
    altitude: Angle
    distance: Au

    def __init__(self, azimuth: float, altitude: float, distance: float):
        self.azimuth = Angle(azimuth)
        self.altitude = Angle(altitude)
        self.distance = Au(distance)

    def house_pos(self) -> (int, float):
        alt = self.altitude.radians()
        azimuth = self.azimuth.radians()
        angle = math.atan2(math.tan(alt), math.cos(azimuth)) / math.pi * 180.0 - 90.0
        if angle < 0:
            angle += 360
        house13 = int(math.floor(angle * 13.0 / 360.0)) + 1
        house_pos = angle * 13.0 / 360.0 + 1.0 - house13
        house_pos *= 180.0 / math.pi
        return house13, house_pos

    def json(self) -> dict:
        return {'azimuth': self.azimuth.degrees, 'alt': self.altitude.degrees, 'dist': self.distance.au}


@dataclass
class EclSpeed(EclCoord):
    """Ecliptic coordinate with speed"""

    longitude_speed: AngularSpeed
    latitude_speed: AngularSpeed
    distance_speed: AuSpeed

    def __init__(self, longitude: float, latitude: float, dist: float, longitude_speed: float, latitude_speed: float, distance_speed: float):
        super().__init__(longitude, latitude, dist)
        self.longitude_speed = AngularSpeed(longitude_speed)
        self.latitude_speed = AngularSpeed(latitude_speed)
        self.distance_speed = AuSpeed(distance_speed)

    def json(self) -> dict:
        d = super().json()
        d['long_spd'] = self.longitude_speed.deg_per_day
        d['lat_spd'] = self.latitude_speed.deg_per_day
        d['dist_spd'] = self.distance_speed.au_per_day
        return d

@dataclass
class EquatorSpeed(EquatorCoord):
    """Ecliptic coordinate with speed"""

    ra_speed: AngularSpeed
    decl_speed: AngularSpeed
    distance_speed: AuSpeed

    def __init__(self, ra: float, decl: float, dist: float, ra_speed: float, decl_speed: float, distance_speed: float):
        super().__init__(ra, decl, dist)
        self.ra_speed = AngularSpeed(ra_speed)
        self.decl_speed = AngularSpeed(decl_speed)
        self.distance_speed = AuSpeed(distance_speed)

    def json(self) -> dict:
        d = super().json()
        d['ra_spd'] = self.ra_speed.deg_per_day
        d['decl_spd'] = self.decl_speed.deg_per_day
        d['dist_spd'] = self.distance_speed.au_per_day
        return d

@dataclass
class BarySpeed(BaryCoord):
    """Barycentric coordinate with speed"""

    longitude_speed: AngularSpeed
    latitude_speed: AngularSpeed
    distance_speed: AuSpeed

    def __init__(self, longitude: float, latitude: float, dist: float, longitude_speed: float, latitude_speed: float, distance_speed: float):
        super().__init__(longitude, latitude, dist)
        self.longitude_speed = AngularSpeed(longitude_speed)
        self.latitude_speed = AngularSpeed(latitude_speed)
        self.distance_speed = AuSpeed(distance_speed)

    def json(self) -> dict:
        d = super().json()
        d['long_spd'] = self.longitude_speed.deg_per_day
        d['lat_spd'] = self.latitude_speed.deg_per_day
        d['dist_spd'] = self.distance_speed.au_per_day
        return d

@dataclass
class HelioSpeed(HelioCoord):
    """Heliocentric coordinate with speed"""

    longitude_speed: AngularSpeed
    latitude_speed: AngularSpeed
    distance_speed: AuSpeed

    def __init__(self, longitude: float, latitude: float, dist: float, longitude_speed: float, latitude_speed: float, distance_speed: float):
        super().__init__(longitude, latitude, dist)
        self.longitude_speed = AngularSpeed(longitude_speed)
        self.latitude_speed = AngularSpeed(latitude_speed)
        self.distance_speed = AuSpeed(distance_speed)

    def json(self) -> dict:
        d = super().json()
        d['long_spd'] = self.longitude_speed.deg_per_day
        d['lat_spd'] = self.latitude_speed.deg_per_day
        d['dist_spd'] = self.distance_speed.au_per_day
        return d



from abc import ABC, abstractmethod
from datetime import datetime, time, timedelta
from types import NoneType

import swisseph as swe

from . import GeoLocation, EclCoord, EquatorCoord, HorCoord, BaryCoord, HelioCoord, EclSpeed, EquatorSpeed, BarySpeed, HelioSpeed

# TODO: Consider using SEFLG_TRUEPOS everywhere (since focal points and apsides use it anyway)

class Celestial(ABC):
    """Abstract class for celestial objects whose location can be computed"""

    NAMES = {
        "SUN": swe.SUN,
        "MOON": swe.MOON,
        "MERCURY": swe.MERCURY,
        "VENUS": swe.VENUS,
        "MARS": swe.MARS,
        "EARTH": swe.EARTH,
        "JUPITER": swe.JUPITER,
        "SATURN": swe.SATURN,
        "URANUS": swe.URANUS,
        "NEPTUNE": swe.NEPTUNE,
        "PLUTO": swe.PLUTO,
        "ERIS": swe.AST_OFFSET + 136199,
        # "QUAOAR": swe.AST_OFFSET + 50000,
        "CETO": swe.AST_OFFSET + 65489,
        "SEDNA": swe.AST_OFFSET + 90377,
    }

    def __init__(self):
        self.name = None

    def ecl_coord(self, time: datetime, location: GeoLocation, *, speed: bool = False, mean: bool = False) -> EclCoord | EclSpeed:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_ecl_coord(jd, speed=speed, mean=mean)

    def equator_coord(self, time: datetime, location: GeoLocation, *, speed: bool = False, mean: bool = False) -> EquatorCoord | EquatorSpeed:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_equator_coord(jd, speed=speed, mean=mean)

    def hor_coord(self, time: datetime, location: GeoLocation, *, mean: bool = False) -> HorCoord:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        coord = self.swe_equator_coord(jd, mean=mean)
        geopos = (location.longitude.degrees, location.latitude.degrees, 0.0)
        pos = (coord.ra.degrees, coord.decl.degrees, 0.0)
        atpress = 0
        attemp = 0
        (azimuth, true_alt, app_alt) = swe.azalt(jd, swe.EQU2HOR, geopos, atpress, attemp, pos)
        return HorCoord(azimuth, true_alt)

    def ecl_speed(self, time: datetime, location: GeoLocation, **kwargs) -> EclSpeed:
        return self.ecl_coord(time, location, speed=True, **kwargs)

    def equator_speed(self, time: datetime, location: GeoLocation, **kwargs) -> EquatorSpeed:
        return self.equator_coord(time, location, speed=True, **kwargs)

    def transits(self, time: datetime, location: GeoLocation):
        return {
            'rise': self.rises(time, location),
            'set': self.sets(time, location),
            'mc': self.mc_trans(time, location),
            'ic': self.ic_trans(time, location),
        }

    def rises(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_RISE)

    def sets(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_SET)

    def mc_trans(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_MTRANSIT)

    def ic_trans(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_ITRANSIT)

    def __rise_trans(self, when: datetime, location: GeoLocation, rsmi: int):
        if self.is_focal_point():
            return None
        tjdut = swe.julday(when.year, when.month, when.day, 0.)
        flags = swe.FLG_SWIEPH | swe.FLG_TOPOCTR
        rsmi |= swe.BIT_DISC_CENTER | swe.BIT_FIXED_DISC_SIZE | swe.BIT_NO_REFRACTION | swe.BIT_ASTRO_TWILIGHT
        lon = location.longitude.degrees
        lat = location.latitude.degrees
        alt = 0.0
        atpress = 0
        attemp = 0
        (found, (jultime, _, _, _, _, _, _, _, _, _)) = swe.rise_trans(
            tjdut, self.swe_id(), rsmi, (lon, lat, alt), atpress, attemp, flags
        )
        if found != 0:
            return None
        (year, month, day, tm) = swe.revjul(jultime)
        if year != when.year or month != when.month or day != when.day:
            return None
        hours = int(tm)
        minutes = (tm - hours) * 60.
        seconds = (tm - hours - int(minutes) / 60) * 60 * 60
        return timedelta(hours=hours, minutes=int(minutes), seconds=int(seconds))

    @abstractmethod
    def swe_id(self):
        pass

    @classmethod
    def swe_id_by_name(cls, name: str) -> int:
        swe_code = cls.NAMES.get(name.upper())
        if swe_code is None:
            raise Exception(f"Unknown planet {name}")
        return swe_code

    @abstractmethod
    def is_fixed(self) -> bool:
        pass

    @abstractmethod
    def is_focal_point(self) -> bool:
        pass

    @abstractmethod
    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord | EclSpeed:
        pass

    @abstractmethod
    def swe_equator_coord(self, jd, speed: bool = False, mean: bool = False) -> EquatorCoord | EclSpeed:
        pass


class Planet(Celestial):
    """Planets (moving physical bodies)"""

    def __init__(self, name: str, swe_code: int | NoneType = None):
        self.name = name
        self.__swe_code = swe_code or Celestial.swe_id_by_name(name)

    def is_fixed(self) -> bool:
        return False

    def is_focal_point(self) -> bool:
        return False

    def bary_coord(self, time: datetime, *, speed: bool = False, mean: bool = False) -> BaryCoord | BarySpeed:
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_bary_coord(jd, speed=speed, mean=mean)

    def helio_coord(self, time: datetime, *, speed: bool = False, mean: bool = False) -> HelioCoord | HelioSpeed:
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_helio_coord(jd, speed=speed, mean=mean)

    def bary_speed(self, time: datetime, **kwargs) -> BarySpeed:
        return self.bary_coord(time, speed=True, **kwargs)

    def helio_speed(self, time: datetime, **kwargs) -> HelioSpeed:
        return self.helio_coord(time, speed=True, **kwargs)

    def swe_id(self):
        return self.__swe_code

    def swe_calc(self, jd, iflag, *, speed: bool = False, mean: bool = False):
        if mean is not False:
            raise RuntimeError("mean position flag has no meaning for the planets")
        iflag |= swe.FLG_SWIEPH
        if speed:
            iflag |= swe.FLG_SPEED
        return swe.calc_ut(jd, self.__swe_code, iflag)

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord | EclSpeed:
        (ecl, _) = self.swe_calc(jd, swe.FLG_TOPOCTR, speed=speed, mean=mean)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord | EquatorSpeed:
        (equator, _) = self.swe_calc(jd, swe.FLG_TOPOCTR | swe.FLG_EQUATORIAL, speed=speed, mean=mean)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])

    def swe_bary_coord(self, jd, speed: bool = False, mean: bool = False) -> BaryCoord | BarySpeed:
        (ecl, _) = self.swe_calc(jd, swe.FLG_BARYCTR, speed=speed, mean=mean)
        if speed:
            return BarySpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4])
        else:
            return BaryCoord(ecl[0], ecl[1], ecl[2])

    def swe_helio_coord(self, jd, speed: bool = False, mean: bool = False) -> HelioCoord | HelioSpeed:
        (ecl, _) = self.swe_calc(jd, swe.FLG_HELCTR, speed=speed, mean=mean)
        if speed:
            return HelioSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4])
        else:
            return HelioCoord(ecl[0], ecl[1], ecl[2])


class ApsisNode(Celestial):
    """Planet apsides and nodes"""

    def __init__(self, name: str, swe_code: int | NoneType = None):
        self.name = name
        self.__swe_code = swe_code or Celestial.swe_id_by_name(name)

    def _swe_ecl_coord_nod_aps(self, jd, *, speed: bool = False, mean: bool = False, equatorial: bool = False, second_focus: bool = False):
        method = swe.NODBIT_MEAN if mean else swe.NODBIT_OSCU
        if second_focus:
            method |= swe.NODBIT_FOPOINT
        iflag = swe.FLG_SWIEPH | swe.FLG_TOPOCTR
        if speed:
            iflag |= swe.FLG_SPEED
        if equatorial:
            iflag |= swe.FLG_EQUATORIAL
        return swe.nod_aps_ut(jd, self.__swe_code, method, iflag)

    def swe_id(self) -> int:
        return self.__swe_code

    def is_fixed(self) -> bool:
        return False

    def is_focal_point(self) -> bool:
        return True


class SecondFocus(ApsisNode):
    """Second focal point of some planet orbit"""

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        (_, _, _, ecl) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=False, second_focus=True)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        (_, _, _, equator) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=True, second_focus=True)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


class ApoApsis(ApsisNode):
    """Aphelion/apogee of some planet orbit"""

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        (_, _, _, ecl) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=False)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        (_, _, _, equator) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=True)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


class PeriApsis(ApsisNode):
    """Perihelion/perigee of some planet orbit"""

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        (_, _, ecl, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=False)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        (_, _, equator, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=True)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


class AscNode(ApsisNode):
    """Ascending node of some planet orbit"""

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        (ecl, _, _, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=False)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        (equator, _, _, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=True)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


class DscNode(ApsisNode):
    """Descending node of some planet orbit"""

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        (_, ecl, _, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=False)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        (_, equator, _, _) = super()._swe_ecl_coord_nod_aps(jd, speed=speed, mean=mean, equatorial=True)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


class FixedCelestial(Celestial):
    """Fixed astronomical objects: stars, galactic and deep space objects"""

    def __init__(self, name: str, swe_code: str):
        self.name = name
        self.__swe_code = swe_code

    def swe_id(self):
        return self.__swe_code

    def is_fixed(self) -> bool:
        return True

    def is_focal_point(self) -> bool:
        return False

    def swe_ecl_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EclCoord:
        if mean is not False:
            raise RuntimeError("mean position flag has no meaning for fixed objects")
        iflag = swe.FLG_SWIEPH | swe.FLG_TOPOCTR
        if speed:
            iflag |= swe.FLG_SPEED
        (ecl, _, _) = swe.fixstar_ut(self.__swe_code, jd, iflag)
        if speed:
            return EclSpeed(ecl[0], ecl[1], ecl[2], ecl[3], ecl[4], ecl[5])
        else:
            return EclCoord(ecl[0], ecl[1], ecl[2])

    def swe_equator_coord(self, jd, *, speed: bool = False, mean: bool = False) -> EquatorCoord:
        if mean is not False:
            raise RuntimeError("mean position flag has no meaning for fixed objects")
        iflag = swe.FLG_SWIEPH | swe.FLG_TOPOCTR | swe.FLG_EQUATORIAL
        if speed:
            iflag |= swe.FLG_SPEED
        (equator, _, _) = swe.fixstar_ut(self.__swe_code, jd, iflag)
        if speed:
            return EquatorSpeed(equator[0], equator[1], equator[2], equator[3], equator[4], equator[5])
        else:
            return EquatorCoord(equator[0], equator[1], equator[2])


Planet.Sun = Planet("Sun")
Planet.Moon = Planet("Moon")
Planet.Mercury = Planet("Mercury")
Planet.Venus = Planet("Venus")
Planet.Earth = Planet("Earth")
Planet.Mars = Planet("Mars")
Planet.Jupiter = Planet("Jupiter")
Planet.Saturn = Planet("Saturn")
Planet.Uranus = Planet("Uranus")
Planet.Neptune = Planet("Neptune")
Planet.Pluto = Planet("Pluto")
Planet.Eris = Planet("Eris")
Planet.Ceto = Planet("Ceto")
Planet.Sedna = Planet("Sedna")

Planet.septener = [Planet.Sun, Planet.Mars, Planet.Moon, Planet.Mercury, Planet.Jupiter, Planet.Venus, Planet.Saturn]
Planet.novile = Planet.septener + [Planet.Uranus, Planet.Neptune]

SecondFocus.Moon = SecondFocus("BE Moon", swe.MOON)
SecondFocus.Mercury = SecondFocus("BS Mercury", swe.MERCURY)
SecondFocus.Venus = SecondFocus("BS Venus", swe.VENUS)
SecondFocus.Earth = SecondFocus("BS Earth", swe.EARTH)
SecondFocus.Mars = SecondFocus("BS Mars", swe.MARS)
SecondFocus.Jupiter = SecondFocus("BS Jupiter", swe.JUPITER)
SecondFocus.Saturn = SecondFocus("BS Saturn", swe.SATURN)
SecondFocus.Uranus = SecondFocus("BS Uranus", swe.URANUS)
SecondFocus.Neptune = SecondFocus("BS Neptune", swe.NEPTUNE)



from .primitives import GeoLocation, Angle, AngularSpeed, Au, AuSpeed
from .coords import HorCoord, EclCoord, EquatorCoord, BaryCoord, HelioCoord, EclSpeed, EquatorSpeed, BarySpeed, HelioSpeed
from .zodiac import Zodiac, ZodiacConstell
from .natal import NatalObject, Natal
from .celestials import Celestial, Planet, ApsisNode, ApoApsis, PeriApsis, AscNode, DscNode, SecondFocus, FixedCelestial

__all__ = [GeoLocation, Angle, AngularSpeed,
           HorCoord, EclCoord, EquatorCoord, BaryCoord, HelioCoord, EclSpeed, EquatorSpeed, BarySpeed, HelioSpeed,
           Zodiac, ZodiacConstell,
           Natal, NatalObject,
           Celestial, Planet, SecondFocus, ApsisNode, ApoApsis, PeriApsis, AscNode, DscNode, FixedCelestial]


from dataclasses import dataclass
from types import NoneType

from dms2dec.dms_convert import dms2dec
import math

@dataclass
class Au:
    """Distance in astronomic units (a.u.)"""

    def __init__(self, au: float):
        self.au = au

@dataclass
class AuSpeed:
    """Speed in astronomic units (a.u.) per day"""

    def __init__(self, speed: float):
        self.au_per_day = speed

@dataclass
class Angle:
    """Angle measured in degrees"""

    def __init__(self, degrees: str | float | int | NoneType = None, *, radians: float | int | NoneType = None):
        if radians is not None and degrees is not None:
            raise RuntimeError("either degrees or radians has to be provided to the method")
        elif degrees is not None:
            if type(degrees) is str:
                self.degrees = dms2dec(degrees)
            elif type(degrees) is float or type(degrees) is int:
                self.degrees = degrees
        elif radians is not None:
            self.degrees = radians * 180.0 / math.pi

    def __eq__(self, other) -> bool:
        return self.degrees == other.degrees

    def radians(self):
        return self.degrees * 2.0 * math.pi / 360.0

    def aspect(self, orb: float = 0.5):
        if -orb <= self.degrees <= orb:
            return 1
        for div in range(2, 14):
            angle = 360.0 / div
            if angle - orb <= self.degrees <= angle + orb:
                return div
        return None


@dataclass
class AngularSpeed:
    """Angular speed measured in degrees per day"""
    def __init__(self, speed: float | int):
        self.deg_per_day = speed


@dataclass
class GeoLocation:
    """Location on the earth"""

    def __init__(self, longitude, latitude):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)



from datetime import datetime, time, timedelta
import swisseph as swe

from .celestials import Celestial
from .primitives import GeoLocation
from .coords import HorCoord, EclCoord, EquatorCoord, EclSpeed
from .zodiac import Zodiac, ZodiacConstell


class NatalObject:
    """Natal object computable type"""

    def __init__(self, obj: Celestial, birth: datetime, place: GeoLocation):
        self.name = obj.name
        self.obj = obj
        self.birth = birth
        self.place = place
        self.__ecl_coord = None
        self.__equator_coord = None
        self.__transits = None

    def julday(self) -> float:
        return swe.julday(self.birth.year, self.birth.month, self.birth.day, self.birth.hour + self.birth.minute / 60.)

    def ecl_coord(self, *, speed: bool = False, mean: bool = False) -> EclCoord | EclSpeed:
        if self.__ecl_coord is None:
            swe.set_topo(self.place.longitude.degrees, self.place.latitude.degrees)
            self.__ecl_coord = self.obj.swe_ecl_coord(self.julday(), speed=speed, mean=mean)
        return self.__ecl_coord

    def equator_coord(self, *, speed: bool = False, mean: bool = False) -> EquatorCoord | EquatorCoord:
        if self.__equator_coord is None:
            swe.set_topo(self.place.longitude.degrees, self.place.latitude.degrees)
            self.__equator_coord = self.obj.swe_equator_coord(self.julday(), speed=speed, mean=mean)
        return self.__equator_coord

    def hor_coord(self) -> HorCoord:
        coord = self.ecl_coord()
        geopos = (self.place.longitude.degrees, self.place.latitude.degrees, 0.0)
        pos = (coord.longitude.degrees, coord.latitude.degrees, 0.0)
        atpress = 0
        attemp = 0
        (azimuth, true_alt, app_alt) = swe.azalt(self.julday(), swe.ECL2HOR, geopos, atpress, attemp, pos)
        return HorCoord(azimuth, true_alt)

    def sign_pos(self) -> (Zodiac, float):
        return self.ecl_coord().sign_pos()

    def constell_pos(self) -> (ZodiacConstell, float):
        return self.ecl_coord().constell_pos()

    def house_pos(self) -> (int, float):
        return self.hor_coord().house_pos()

    def transits(self) -> dict:
        if self.__transits is None:
            self.__transits = self.obj.transits(self.birth, self.place)
        return self.__transits


class Natal:
    """Natal chart"""

    def __init__(self, person: str, birth: datetime, place: GeoLocation, celestials: [Celestial]):
        self.person = person
        self.birth = birth
        self.place = place
        self.celestials = {obj: NatalObject(obj, birth, place) for obj in celestials}

    def __iter__(self):
        for cel in self.celestials:
            yield self.celestials[cel]

    def __getitem__(self, celestial: Celestial) -> NatalObject:
        return self.celestials[celestial]

    def aspects(self, orb: float = 1.01, of=None, to=None):
        if type(of) is list:
            list1 = of
        elif isinstance(of, Celestial):
            list1 = [of]
        else:
            list1 = self.celestials

        if type(to) is list:
            list2 = to
        elif isinstance(to, Celestial):
            list2 = [to]
        else:
            list2 = self.celestials

        for (i1, cel1) in enumerate(list1):
            for (i2, cel2) in enumerate(list2):
                if cel1 == cel2:
                    break
                angle = self[cel1].ecl_coord() ^ self[cel2].ecl_coord()
                aspect = angle.aspect(orb=orb)
                if aspect is None:
                    continue
                yield {"first": cel1.name, "second": cel2.name, "aspect": aspect}

    def parans(self, orb: timedelta = timedelta(minutes=5)) -> list:
        events = []
        timeline = {}
        for cel in self:
            transits = cel.transits()
            for transit_name in transits:
                transit_time = transits[transit_name]
                if transit_time is None:
                    continue
                for tm in timeline:
                    if abs(tm - transit_time) <= orb:
                        events.append([
                            timeline[tm],
                            (transit_name, cel.obj)
                        ])
                timeline[transit_time] = (transit_name, cel.obj)
        return events
